"""Audio output to speakers."""

import asyncio
import contextlib
import logging
from asyncio.subprocess import Process
from dataclasses import dataclass
from typing import List, Optional

from .audio import AudioChunk, AudioChunkConverter
from .client import AsyncClient
from .event import Event, Eventable

_LOGGER = logging.getLogger(__name__)

_PLAYED_TYPE = "played"


@dataclass
class Played(Eventable):
    """Audio has finished playing."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _PLAYED_TYPE

    def event(self) -> Event:
        return Event(type=_PLAYED_TYPE)

    @staticmethod
    def from_event(event: Event) -> "Played":
        return Played()


class SndProcessAsyncClient(AsyncClient, contextlib.AbstractAsyncContextManager):
    """Context manager for sending output audio to an external program."""

    def __init__(
        self,
        rate: int,
        width: int,
        channels: int,
        program: str,
        program_args: List[str],
    ) -> None:
        super().__init__()

        self.rate = rate
        self.width = width
        self.channels = channels
        self.program = program
        self.program_args = program_args

        self._proc: Optional[Process] = None
        self._chunk_converter = AudioChunkConverter(rate, width, channels)

    async def connect(self) -> None:
        self._proc = await asyncio.create_subprocess_exec(
            self.program, *self.program_args, stdin=asyncio.subprocess.PIPE
        )

    async def disconnect(self) -> None:
        assert self._proc is not None
        assert self._proc.stdin is not None

        try:
            if self._proc.returncode is None:
                # Terminate process gracefully
                self._proc.stdin.close()
                await self._proc.wait()
        except ProcessLookupError:
            # Expected when process has already exited
            pass
        except Exception:
            _LOGGER.exception("Unexpected error stopping process: %s", self.program)
        finally:
            self._proc = None

    async def __aenter__(self) -> "SndProcessAsyncClient":
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.disconnect()

    async def read_event(self) -> Optional[Event]:
        """Client is write-only."""

    async def write_event(self, event: Event) -> None:
        assert self._proc is not None
        assert self._proc.stdin is not None

        if not AudioChunk.is_type(event.type):
            return

        chunk = AudioChunk.from_event(event)

        # Convert sample rate/width/channels if necessary
        chunk = self._chunk_converter.convert(chunk)

        self._proc.stdin.write(chunk.audio)
        await self._proc.stdin.drain()



"""Error event."""

from dataclasses import dataclass
from typing import Any, Dict, Optional

from .event import Event, Eventable

_ERROR_TYPE = "error"


@dataclass
class Error(Eventable):
    """Error with text and an optional code."""

    text: str
    """Human-readable error message."""

    code: Optional[str] = None
    """Machine-readable error code."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _ERROR_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {"text": self.text}
        if self.code is not None:
            data["code"] = self.code

        return Event(type=_ERROR_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "Error":
        return Error(text=event.data["text"], code=event.data.get("code"))



"""Server tests."""

import asyncio
import socket
import tempfile
from pathlib import Path

import pytest

from wyoming.client import AsyncClient
from wyoming.event import Event
from wyoming.ping import Ping, Pong
from wyoming.server import (
    AsyncEventHandler,
    AsyncServer,
    AsyncStdioServer,
    AsyncTcpServer,
    AsyncUnixServer,
)


class PingHandler(AsyncEventHandler):
    async def handle_event(self, event: Event) -> bool:
        if Ping.is_type(event.type):
            ping = Ping.from_event(event)
            await self.write_event(Pong(text=ping.text).event())
            return False

        return True


def test_from_uri() -> None:
    """Test AsyncServer.from_uri"""
    # Bad scheme
    with pytest.raises(ValueError):
        AsyncServer.from_uri("ftp://127.0.0.1:5000")

    # Missing hostname
    with pytest.raises(ValueError):
        AsyncServer.from_uri("tcp://:5000")

    # Missing port
    with pytest.raises(ValueError):
        AsyncServer.from_uri("tcp://127.0.0.1")

    stdio_server = AsyncServer.from_uri("stdio://")
    assert isinstance(stdio_server, AsyncStdioServer)

    tcp_server = AsyncServer.from_uri("tcp://127.0.0.1:5000")
    assert isinstance(tcp_server, AsyncTcpServer)
    assert tcp_server.host == "127.0.0.1"
    assert tcp_server.port == 5000

    unix_server = AsyncServer.from_uri("unix:///path/to/socket")
    assert isinstance(unix_server, AsyncUnixServer)
    assert unix_server.socket_path == Path("/path/to/socket")


@pytest.mark.asyncio
async def test_unix_server() -> None:
    """Test sending events to and from a Unix socket server."""
    with tempfile.TemporaryDirectory() as temp_dir:
        socket_path = Path(temp_dir) / "test.socket"
        uri = f"unix://{socket_path}"
        unix_server = AsyncServer.from_uri(uri)
        await unix_server.start(PingHandler)

        # Wait for path to exist
        while not socket_path.exists():
            await asyncio.sleep(0.1)

        client = AsyncClient.from_uri(uri)
        await client.connect()
        await client.write_event(Ping(text="test").event())

        event = await asyncio.wait_for(client.read_event(), timeout=1)
        assert event is not None
        assert Pong.is_type(event.type)
        assert Pong.from_event(event).text == "test"

        await client.disconnect()
        await unix_server.stop()


@pytest.mark.asyncio
async def test_tcp_server() -> None:
    """Test sending events to and from a TCP server."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    port = sock.getsockname()[1]
    sock.close()

    uri = f"tcp://127.0.0.1:{port}"

    tcp_server = AsyncServer.from_uri(uri)
    await tcp_server.start(PingHandler)

    # Wait for socket to open
    for _ in range(10):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(("127.0.0.1", port))
            break
        except ConnectionRefusedError:
            await asyncio.sleep(0.1)

    client = AsyncClient.from_uri(uri)
    await client.connect()
    await client.write_event(Ping(text="test").event())

    event = await asyncio.wait_for(client.read_event(), timeout=1)
    assert event is not None
    assert Pong.is_type(event.type)
    assert Pong.from_event(event).text == "test"

    await client.disconnect()
    await tcp_server.stop()



"""Information about available services, models, etc.."""

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from .audio import AudioFormat
from .event import Event, Eventable
from .util.dataclasses_json import DataClassJsonMixin

DOMAIN = "info"
_DESCRIBE_TYPE = "describe"
_INFO_TYPE = "info"
_SELECT_PROGRAM_TYPE = "select-program"


@dataclass
class Describe(Eventable):
    """Request info message."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _DESCRIBE_TYPE

    def event(self) -> Event:
        return Event(type=_DESCRIBE_TYPE)

    @staticmethod
    def from_event(event: Event) -> "Describe":
        return Describe()


@dataclass
class SelectProgram(Eventable):
    """Select which program handles this connection.

    Sent by the client after connecting, before the first request event (e.g.,
    describe/transcribe/synthesize/detect/recognize). Selects one of the
    programs advertised in the info message by name for the lifetime of the
    connection. The domain (asr, tts, ...) is implied by the request events
    that follow, so ``name`` only needs to be unique within a domain.

    If this event is not sent, the first program of each type in the info
    message is used. Servers are expected to drop unrecognized events, so
    sending this to a server that predates it is a no-op (the default program
    is used).
    """

    name: str
    """Name of the program to use (matches a program name in the info message)."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _SELECT_PROGRAM_TYPE

    def event(self) -> Event:
        return Event(type=_SELECT_PROGRAM_TYPE, data={"name": self.name})

    @staticmethod
    def from_event(event: Event) -> "SelectProgram":
        return SelectProgram(name=event.data["name"])


@dataclass
class Attribution(DataClassJsonMixin):
    """Attribution for an artifact."""

    name: str
    """Who made it."""

    url: str
    """Where it's from."""


@dataclass
class Artifact(DataClassJsonMixin):
    """Information about a service, model, etc.."""

    name: str
    """Name/id of artifact."""

    attribution: Attribution
    """Who made the artifact and where it's from."""

    installed: bool
    """True if the artifact is currently installed."""

    description: Optional[str]
    """Human-readable description of the artifact."""

    version: Optional[str]
    """Version of the artifact."""


# -----------------------------------------------------------------------------


@dataclass
class AsrModel(Artifact):
    """Speech-to-text model."""

    languages: List[str]
    """List of supported model languages."""


@dataclass
class AsrProgram(Artifact):
    """Speech-to-text service."""

    models: List[AsrModel]
    """List of available models."""

    supports_transcript_streaming: bool = False
    """True if transcript streaming events are supported."""

    requires_external_vad: bool = True
    """True if ASR program needs an external VAD to detect the end of voice commands."""

    prefers_auto_gain_enabled: bool = True
    """True if input audio should adjust gain automatically for best results."""

    prefers_noise_reduction_enabled: bool = True
    """True if input audio should apply noise reduction for best results."""


# -----------------------------------------------------------------------------


@dataclass
class TtsVoiceSpeaker(DataClassJsonMixin):
    """Individual speaker in a multi-speaker voice."""

    name: str
    """Name/id of speaker."""


@dataclass
class TtsVoice(Artifact):
    """Text-to-speech voice."""

    languages: List[str]
    """List of languages available in the voice."""

    speakers: Optional[List[TtsVoiceSpeaker]] = None
    """List of individual speakers in the voice."""


@dataclass
class TtsProgram(Artifact):
    """Text-to-speech service."""

    voices: List[TtsVoice]
    """List of available voices."""

    supports_synthesize_streaming: bool = False
    """True if synthesize text streaming events are supported."""


# -----------------------------------------------------------------------------


@dataclass
class HandleModel(Artifact):
    """Intent handling model."""

    languages: List[str]
    """List of supported languages in the model."""


@dataclass
class HandleProgram(Artifact):
    """Intent handling service."""

    models: List[HandleModel]
    """List of available models."""

    supports_handled_streaming: bool = False
    """True if handled response streaming events are supported."""

    supports_home_control: bool = False
    """True if the service may execute home control actions during handling."""


# -----------------------------------------------------------------------------


@dataclass
class WakeModel(Artifact):
    """Wake word detection model."""

    languages: List[str]
    """List of languages supported by the model."""

    phrase: Optional[str]
    """Wake up phrase used by the model."""


@dataclass
class WakeProgram(Artifact):
    """Wake word detection service."""

    models: List[WakeModel]
    """List of available models."""


# -----------------------------------------------------------------------------


@dataclass
class IntentModel(Artifact):
    """Intent recognition model."""

    languages: List[str]
    """List of languages supported by the model."""


@dataclass
class IntentProgram(Artifact):
    """Intent recognition service."""

    models: List[IntentModel]
    """List of available models."""


# -----------------------------------------------------------------------------


@dataclass
class Satellite(Artifact):
    """Satellite information."""

    area: Optional[str] = None
    """Name of the area the satellite is in."""

    has_vad: Optional[bool] = None
    """True if a local VAD will be used to detect the end of voice commands."""

    active_wake_words: Optional[List[str]] = None
    """Wake words that are currently being listened for."""

    max_active_wake_words: Optional[int] = None
    """Maximum number of local wake words that can be run simultaneously."""

    supports_trigger: Optional[bool] = None
    """Satellite supports remotely triggering pipeline runs."""


# -----------------------------------------------------------------------------


@dataclass
class MicProgram(Artifact):
    """Microphone information."""

    mic_format: AudioFormat
    """Input audio format."""


# -----------------------------------------------------------------------------


@dataclass
class SndProgram(Artifact):
    """Sound output information."""

    snd_format: AudioFormat
    """Output audio format."""


# -----------------------------------------------------------------------------


@dataclass
class Info(Eventable):
    """Response to describe message with information about available services, models, etc."""

    asr: List[AsrProgram] = field(default_factory=list)
    """Speech-to-text services."""

    tts: List[TtsProgram] = field(default_factory=list)
    """Text-to-speech services."""

    handle: List[HandleProgram] = field(default_factory=list)
    """Intent handling services."""

    intent: List[IntentProgram] = field(default_factory=list)
    """Intent recognition services."""

    wake: List[WakeProgram] = field(default_factory=list)
    """Wake word detection services."""

    mic: List[MicProgram] = field(default_factory=list)
    """Audio input services."""

    snd: List[SndProgram] = field(default_factory=list)
    """Audio output services."""

    satellite: Optional[Satellite] = None
    """Satellite information."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _INFO_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {
            "asr": [p.to_dict() for p in self.asr],
            "tts": [p.to_dict() for p in self.tts],
            "handle": [p.to_dict() for p in self.handle],
            "intent": [p.to_dict() for p in self.intent],
            "wake": [p.to_dict() for p in self.wake],
            "mic": [p.to_dict() for p in self.mic],
            "snd": [p.to_dict() for p in self.snd],
        }

        if self.satellite is not None:
            data["satellite"] = self.satellite.to_dict()

        return Event(type=_INFO_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "Info":
        satellite: Optional[Satellite] = None
        satellite_data = event.data.get("satellite")
        if satellite_data is not None:
            satellite = Satellite.from_dict(satellite_data)

        return Info(
            asr=[AsrProgram.from_dict(d) for d in event.data.get("asr", [])],
            tts=[TtsProgram.from_dict(d) for d in event.data.get("tts", [])],
            handle=[HandleProgram.from_dict(d) for d in event.data.get("handle", [])],
            intent=[IntentProgram.from_dict(d) for d in event.data.get("intent", [])],
            wake=[WakeProgram.from_dict(d) for d in event.data.get("wake", [])],
            mic=[MicProgram.from_dict(d) for d in event.data.get("mic", [])],
            snd=[SndProgram.from_dict(d) for d in event.data.get("snd", [])],
            satellite=satellite,
        )



"""Audio input/output."""

import argparse
import io
import sys
import wave
from dataclasses import dataclass
from typing import Iterable, Optional, Union

try:
    # Use built-in audioop until it's removed in Python 3.13
    import audioop  # pylint: disable=deprecated-module
except ImportError:
    from . import pyaudioop as audioop  # type: ignore[no-redef]

from .event import Event, Eventable
from .util.dataclasses_json import DataClassJsonMixin

_CHUNK_TYPE = "audio-chunk"
_START_TYPE = "audio-start"
_STOP_TYPE = "audio-stop"


@dataclass
class AudioFormat(DataClassJsonMixin):
    """Base class for events with audio format information."""

    rate: int
    """Hertz"""

    width: int
    """Bytes"""

    channels: int
    """Mono = 1"""


@dataclass
class AudioChunk(AudioFormat, Eventable):
    """Chunk of raw PCM audio."""

    audio: bytes
    """Raw audio"""

    timestamp: Optional[int] = None
    """Milliseconds"""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _CHUNK_TYPE

    def event(self) -> Event:
        return Event(
            type=_CHUNK_TYPE,
            data={
                "rate": self.rate,
                "width": self.width,
                "channels": self.channels,
                "timestamp": self.timestamp,
            },
            payload=self.audio,
        )

    @staticmethod
    def from_event(event: Event) -> "AudioChunk":
        return AudioChunk(
            rate=event.data["rate"],
            width=event.data["width"],
            channels=event.data["channels"],
            audio=event.payload or bytes(),
            timestamp=event.data.get("timestamp"),
        )

    @property
    def samples(self) -> int:
        return len(self.audio) // (self.width * self.channels)

    @property
    def seconds(self) -> float:
        return self.samples / self.rate

    @property
    def milliseconds(self) -> int:
        return int(self.seconds * 1_000)


@dataclass
class AudioStart(AudioFormat, Eventable):
    """Audio stream has started."""

    timestamp: Optional[int] = None
    """Milliseconds"""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _START_TYPE

    def event(self) -> Event:
        return Event(
            type=_START_TYPE,
            data={
                "rate": self.rate,
                "width": self.width,
                "channels": self.channels,
                "timestamp": self.timestamp,
            },
        )

    @staticmethod
    def from_event(event: Event) -> "AudioStart":
        return AudioStart(
            rate=event.data["rate"],
            width=event.data["width"],
            channels=event.data["channels"],
            timestamp=event.data.get("timestamp"),
        )


@dataclass
class AudioStop(Eventable):
    """Audio stream has stopped."""

    timestamp: Optional[int] = None
    """Milliseconds"""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _STOP_TYPE

    def event(self) -> Event:
        return Event(
            type=_STOP_TYPE,
            data={"timestamp": self.timestamp},
        )

    @staticmethod
    def from_event(event: Event) -> "AudioStop":
        return AudioStop(timestamp=event.data.get("timestamp"))


@dataclass
class AudioChunkConverter:
    """Converts audio chunks using audioop/pyaudioop."""

    rate: Optional[int] = None
    width: Optional[int] = None
    channels: Optional[int] = None
    _ratecv_state = None

    def convert(self, chunk: AudioChunk) -> AudioChunk:
        """Converts sample rate, width, and channels as necessary."""
        if (
            ((self.rate is None) or (chunk.rate == self.rate))
            and ((self.width is None) or (chunk.width == self.width))
            and ((self.channels is None) or (chunk.channels == self.channels))
        ):
            return chunk

        audio_bytes = chunk.audio
        width = chunk.width

        if (self.width is not None) and (chunk.width != self.width):
            # Convert sample width
            audio_bytes = audioop.lin2lin(audio_bytes, chunk.width, self.width)
            width = self.width

        channels = chunk.channels
        if (self.channels is not None) and (chunk.channels != self.channels):
            # Convert to mono or stereo
            if self.channels == 1:
                audio_bytes = audioop.tomono(audio_bytes, width, 1.0, 1.0)
            elif self.channels == 2:
                audio_bytes = audioop.tostereo(audio_bytes, width, 1.0, 1.0)
            else:
                raise ValueError(f"Cannot convert to channels: {self.channels}")

            channels = self.channels

        rate = chunk.rate
        if (self.rate is not None) and (chunk.rate != self.rate):
            # Resample
            audio_bytes, self._ratecv_state = audioop.ratecv(
                audio_bytes,
                width,
                channels,
                chunk.rate,
                self.rate,
                self._ratecv_state,
            )
            rate = self.rate

        return AudioChunk(
            rate,
            width,
            channels,
            audio_bytes,
            timestamp=chunk.timestamp,
        )


def wav_to_chunks(
    wav_file: wave.Wave_read,
    samples_per_chunk: int,
    timestamp: int = 0,
    start_event: bool = False,
    stop_event: bool = False,
) -> Iterable[Union[AudioStart, AudioChunk, AudioStop]]:
    """Splits WAV file into AudioChunks."""
    rate = wav_file.getframerate()
    width = wav_file.getsampwidth()
    channels = wav_file.getnchannels()

    if start_event:
        yield AudioStart(rate=rate, width=width, channels=channels, timestamp=0)

    audio_bytes = wav_file.readframes(samples_per_chunk)
    while audio_bytes:
        chunk = AudioChunk(
            rate=rate,
            width=width,
            channels=channels,
            audio=audio_bytes,
            timestamp=timestamp,
        )
        yield chunk
        timestamp += chunk.milliseconds
        audio_bytes = wav_file.readframes(samples_per_chunk)

    if stop_event:
        yield AudioStop(timestamp=timestamp)


# -----------------------------------------------------------------------------


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--rate", type=int)
    parser.add_argument("--width", type=int)
    parser.add_argument("--channels", type=int)
    parser.add_argument("--samples-per-chunk", type=int, default=1024)
    args = parser.parse_args()

    converter = AudioChunkConverter(
        rate=args.rate, width=args.width, channels=args.channels
    )

    with io.BytesIO(
        sys.stdin.buffer.read()
    ) as input_wav_io, io.BytesIO() as output_wav_io:
        input_wav_file: wave.Wave_read = wave.open(input_wav_io, "rb")
        output_wav_file: wave.Wave_write = wave.open(output_wav_io, "wb")
        with input_wav_file, output_wav_file:
            # Input
            rate = input_wav_file.getframerate()
            width = input_wav_file.getsampwidth()
            channels = input_wav_file.getnchannels()

            # Output
            output_wav_file.setframerate(args.rate if args.rate is not None else rate)
            output_wav_file.setsampwidth(
                args.width if args.width is not None else width
            )
            output_wav_file.setnchannels(
                args.channels if args.channels is not None else channels
            )

            audio_bytes = input_wav_file.readframes(args.samples_per_chunk)
            while audio_bytes:
                chunk = converter.convert(
                    AudioChunk(rate, width, channels, audio_bytes)
                )
                output_wav_file.writeframes(chunk.audio)
                audio_bytes = input_wav_file.readframes(args.samples_per_chunk)

        sys.stdout.buffer.write(output_wav_io.getvalue())


if __name__ == "__main__":
    main()



"""Test audio utilities."""

import io
import wave

from wyoming.audio import AudioChunk, AudioChunkConverter, wav_to_chunks


def test_chunk_converter() -> None:
    """Test audio chunk converter."""
    converter = AudioChunkConverter(rate=16000, width=2, channels=1)
    input_chunk = AudioChunk(
        rate=48000,
        width=4,
        channels=2,
        audio=bytes(1 * 48000 * 4 * 2),  # 1 sec
    )

    output_chunk = converter.convert(input_chunk)
    assert output_chunk.rate == 16000
    assert output_chunk.width == 2
    assert output_chunk.channels == 1
    assert len(output_chunk.audio) == 1 * 16000 * 2 * 1  # 1 sec


def test_wav_to_chunks() -> None:
    """Test WAV file to audio chunks."""
    with io.BytesIO() as wav_io:
        wav_write: wave.Wave_write = wave.open(wav_io, "wb")
        with wav_write:
            wav_write.setframerate(16000)
            wav_write.setsampwidth(2)
            wav_write.setnchannels(1)
            wav_write.writeframes(bytes(1 * 16000 * 2 * 1))  # 1 sec

        wav_io.seek(0)
        wav_bytes = wav_io.getvalue()

    with io.BytesIO(wav_bytes) as wav_io:
        wav_read: wave.Wave_read = wave.open(wav_io, "rb")
        chunks = list(wav_to_chunks(wav_read, samples_per_chunk=1000))
        assert len(chunks) == 16
        for chunk in chunks:
            assert isinstance(chunk, AudioChunk)
            assert chunk.rate == 16000
            assert chunk.width == 2
            assert chunk.channels == 1
            assert len(chunk.audio) == 1000 * 2  # 1000 samples



"""User-defined events."""

from dataclasses import dataclass
from typing import Any, Dict, Optional

from .event import Event, Eventable

_USER_EVENT_TYPE = "user-event"


@dataclass
class UserEvent(Eventable):
    """User-defined event."""

    name: str
    """Name of user event type."""

    data: Optional[Dict[str, Any]] = None
    """User event data."""

    context: Optional[Dict[str, Any]] = None
    """Context from previous interactions."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _USER_EVENT_TYPE

    def event(self) -> Event:
        data: Dict[str, Any] = {"name": self.name}
        if self.data is not None:
            data["data"] = self.data
        if self.context is not None:
            data["context"] = self.context
        return Event(type=_USER_EVENT_TYPE, data=data)

    @staticmethod
    def from_event(event: Event) -> "UserEvent":
        return UserEvent(
            name=event.data["name"],
            data=event.data.get("data"),
            context=event.data.get("context"),
        )



"""Peer-to-peer protocol for voice assistants."""

from .version import __version__

__author__ = "Michael Hansen"
__email__ = "mike@rhasspy.org"

__all__ = ["__author__", "__email__", "__version__"]



"""Microphone input."""

import asyncio
import contextlib
import logging
import time
from asyncio.subprocess import Process
from typing import List, Optional

from .audio import AudioChunk
from .client import AsyncClient
from .event import Event

_LOGGER = logging.getLogger(__name__)

DOMAIN = "mic"


class MicProcessAsyncClient(AsyncClient, contextlib.AbstractAsyncContextManager):
    """Context manager for getting microphone audio from an external program."""

    def __init__(
        self,
        rate: int,
        width: int,
        channels: int,
        samples_per_chunk: int,
        program: str,
        program_args: List[str],
    ) -> None:
        super().__init__()

        self.rate = rate
        self.width = width
        self.channels = channels
        self.samples_per_chunk = samples_per_chunk
        self.bytes_per_chunk = samples_per_chunk * width * channels
        self.program = program
        self.program_args = program_args

        self._proc: Optional[Process] = None

    async def connect(self) -> None:
        self._proc = await asyncio.create_subprocess_exec(
            self.program, *self.program_args, stdout=asyncio.subprocess.PIPE
        )

    async def disconnect(self) -> None:
        assert self._proc is not None

        try:
            if self._proc.returncode is None:
                # Terminate process gracefully
                self._proc.terminate()
                await self._proc.wait()
        except ProcessLookupError:
            # Expected when process has already exited
            pass
        except Exception:
            _LOGGER.exception("Unexpected error stopping process: %s", self.program)
        finally:
            self._proc = None

    async def __aenter__(self) -> "MicProcessAsyncClient":
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.disconnect()

    async def read_event(self) -> Optional[Event]:
        assert self._proc is not None
        assert self._proc.stdout is not None

        try:
            audio_bytes = await self._proc.stdout.readexactly(self.bytes_per_chunk)
            return AudioChunk(
                rate=self.rate,
                width=self.width,
                channels=self.channels,
                audio=audio_bytes,
                timestamp=time.monotonic_ns(),
            ).event()
        except asyncio.IncompleteReadError:
            return None

    async def write_event(self, event: Event) -> None:
        """Client is read-only."""



"""Shared code for HTTP servers."""

import argparse
from pathlib import Path
from typing import Union

from flask import Flask, jsonify, redirect, request
from swagger_ui import flask_api_doc  # pylint: disable=no-name-in-module

from wyoming.client import AsyncClient
from wyoming.info import Describe, Info


def get_argument_parser() -> argparse.ArgumentParser:
    """Create argument parser with shared arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", default="0.0.0.0")
    parser.add_argument("--port", type=int, default=5000)
    parser.add_argument("--uri", help="URI of Wyoming service")
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG logs to console"
    )
    return parser


def get_app(
    name: str, openapi_config_path: Union[str, Path], args: argparse.Namespace
) -> Flask:
    """Create Flask app with default endpoints."""

    app = Flask(name)

    @app.route("/")
    def redirect_to_api():
        return redirect("/api")

    @app.route("/api/info", methods=["GET"])
    async def api_info():
        uri = request.args.get("uri", args.uri)
        if not uri:
            raise ValueError("URI is required")

        async with AsyncClient.from_uri(uri) as client:
            await client.write_event(Describe().event())

            while True:
                event = await client.read_event()
                if event is None:
                    raise RuntimeError("Client disconnected")

                if Info.is_type(event.type):
                    info = Info.from_event(event)
                    return jsonify(info.to_dict())

    @app.errorhandler(Exception)
    async def handle_error(err):
        """Return error as text."""
        return (f"{err.__class__.__name__}: {err}", 500)

    flask_api_doc(
        app, config_path=str(openapi_config_path), url_prefix="/api", title="API doc"
    )

    return app



"""Implement a tiny subset of dataclasses_json for config."""

from collections.abc import Mapping, Sequence
from dataclasses import asdict, fields, is_dataclass
from typing import Any, Dict, Type


class DataClassJsonMixin:
    """Adds from_dict to dataclass."""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> Any:
        """Parse dataclasses recursively."""
        kwargs: Dict[str, Any] = {}

        cls_fields = {field.name: field for field in fields(cls)}  # type: ignore[arg-type]
        for key, value in data.items():
            if key not in cls_fields:
                # Skip unknown fields
                continue

            field = cls_fields[key]
            if is_dataclass(field.type):
                assert issubclass(field.type, DataClassJsonMixin), field.type  # type: ignore[arg-type,union-attr]
                kwargs[key] = field.type.from_dict(value)  # type: ignore[union-attr]
            else:
                kwargs[key] = _decode(value, field.type)  # type: ignore[arg-type]

        # Fill in optional fields with None
        for field in cls_fields.values():
            if (field.name not in kwargs) and _is_optional(field.type):  # type: ignore[arg-type]
                kwargs[field.name] = None

        return cls(**kwargs)

    def to_dict(self) -> Dict[str, Any]:
        """Alias for asdict."""
        return asdict(self)  # type: ignore[call-overload]


def _decode(value: Any, target_type: Type) -> Any:
    """Decode value using (possibly generic) type."""
    if is_dataclass(target_type):
        assert issubclass(target_type, DataClassJsonMixin), target_type
        return target_type.from_dict(value) if value is not None else None

    if hasattr(target_type, "__args__"):
        # Optional[T]
        if type(None) in target_type.__args__:
            optional_type = target_type.__args__[0]
            return _decode(value, optional_type)

        # List[T]
        if isinstance(value, Sequence):
            list_type = target_type.__args__[0]
            return [_decode(item, list_type) for item in value]

        # Dict[str, T]
        if isinstance(value, Mapping):
            value_type = target_type.__args__[1]
            return {
                map_key: _decode(map_value, value_type)
                for map_key, map_value in value.items()
            }

    return value


def _is_optional(target_type: Type):
    """True if type is Optional"""
    return hasattr(target_type, "__args__") and (type(None) in target_type.__args__)



"""Client tests."""

from pathlib import Path

import pytest

from wyoming.client import (
    AsyncClient,
    AsyncStdioClient,
    AsyncTcpClient,
    AsyncUnixClient,
)


def test_from_uri() -> None:
    """Test AsyncClient.from_uri"""
    # Bad scheme
    with pytest.raises(ValueError):
        AsyncClient.from_uri("ftp://127.0.0.1:5000")

    # Missing hostname
    with pytest.raises(ValueError):
        AsyncClient.from_uri("tcp://:5000")

    # Missing port
    with pytest.raises(ValueError):
        AsyncClient.from_uri("tcp://127.0.0.1")

    stdio_client = AsyncClient.from_uri("stdio://")
    assert isinstance(stdio_client, AsyncStdioClient)

    tcp_client = AsyncClient.from_uri("tcp://127.0.0.1:5000")
    assert isinstance(tcp_client, AsyncTcpClient)
    assert tcp_client.host == "127.0.0.1"
    assert tcp_client.port == 5000

    unix_client = AsyncClient.from_uri("unix:///path/to/socket")
    assert isinstance(unix_client, AsyncUnixClient)
    assert unix_client.socket_path == Path("/path/to/socket")



"""Test event reading/writing."""

import io
import json
from typing import Iterable

import pytest

from wyoming import __version__ as wyoming_version
from wyoming.event import (
    Event,
    async_read_event,
    async_write_event,
    read_event,
    write_event,
)

PAYLOAD = b"test\npayload"
DATA = {"test": "data"}
DATA_BYTES = json.dumps(DATA, ensure_ascii=False).encode("utf-8")


class FakeStreamWriter:
    def __init__(self) -> None:
        self._undrained_data = bytes()
        self._value = bytes()

    def write(self, data: bytes) -> None:
        self._undrained_data += data

    def writelines(self, data: Iterable[bytes]) -> None:
        for line in data:
            self.write(line)

    async def drain(self) -> None:
        self._value += self._undrained_data
        self._undrained_data = bytes()

    def getvalue(self) -> bytes:
        return self._value


class FakeStreamReader:
    def __init__(self, value: bytes) -> None:
        self._value_io = io.BytesIO(value)

    async def readline(self) -> bytes:
        return self._value_io.readline()

    async def readexactly(self, n: int) -> bytes:
        data = self._value_io.read(n)
        assert len(data) == n
        return data


# -----------------------------------------------------------------------------


def test_write_event() -> None:
    """Test synchronous event writing."""
    event = Event(type="test-event", data=DATA, payload=PAYLOAD)
    with io.BytesIO() as buf:
        write_event(event, buf)
        buf.seek(0)
        event_bytes = buf.getvalue()

    with io.BytesIO(event_bytes) as reader:
        # First line is JSON with event type and data/payload lengths.
        #
        # A "data" field may also be present here, but it should be short to
        # avoid overflowing a line buffer.
        assert json.loads(reader.readline()) == {
            "type": event.type,
            "version": wyoming_version,
            "data_length": len(DATA_BYTES),
            "payload_length": len(PAYLOAD),
        }

        # Data dict comes next, encoded as UTF-8 JSON.
        # It should be merged on top of the "data" field, if present.
        assert reader.read(len(DATA_BYTES)) == DATA_BYTES

        # Payload comes last
        assert reader.read(len(PAYLOAD)) == PAYLOAD

        # No more data
        assert reader.read(1) == b""


@pytest.mark.asyncio
async def test_async_write_event() -> None:
    """Test asynchronous event writing."""
    event = Event(type="test-event", data=DATA, payload=PAYLOAD)
    writer = FakeStreamWriter()
    await async_write_event(event, writer)  # type: ignore
    event_bytes = writer.getvalue()

    with io.BytesIO(event_bytes) as reader:
        # First line is JSON with event type and data/payload lengths.
        #
        # A "data" field may also be present here, but it should be short to
        # avoid overflowing a line buffer.
        assert json.loads(reader.readline()) == {
            "type": event.type,
            "version": wyoming_version,
            "data_length": len(DATA_BYTES),
            "payload_length": len(PAYLOAD),
        }

        # Data dict comes next, encoded as UTF-8 JSON.
        # It should be merged on top of the "data" field, if present.
        assert reader.read(len(DATA_BYTES)) == DATA_BYTES

        # Payload comes last
        assert reader.read(len(PAYLOAD)) == PAYLOAD

        # No more data
        assert reader.read(1) == b""


def test_write_event_no_payload() -> None:
    """Test synchronous event writing without a payload."""
    event = Event(type="test-event", data=DATA, payload=None)
    with io.BytesIO() as buf:
        write_event(event, buf)
        buf.seek(0)
        event_bytes = buf.getvalue()

    with io.BytesIO(event_bytes) as reader:
        assert json.loads(reader.readline()) == {
            "type": event.type,
            "version": wyoming_version,
            "data_length": len(DATA_BYTES),
        }

        assert reader.read(len(DATA_BYTES)) == DATA_BYTES

        # No payload
        assert reader.read(1) == b""


@pytest.mark.asyncio
async def test_async_write_event_no_payload() -> None:
    """Test asynchronous event writing without a payload."""
    event = Event(type="test-event", data=DATA, payload=None)
    writer = FakeStreamWriter()
    await async_write_event(event, writer)  # type: ignore
    event_bytes = writer.getvalue()

    with io.BytesIO(event_bytes) as reader:
        assert json.loads(reader.readline()) == {
            "type": event.type,
            "version": wyoming_version,
            "data_length": len(DATA_BYTES),
        }

        assert reader.read(len(DATA_BYTES)) == DATA_BYTES

        # No payload
        assert reader.read(1) == b""


def test_read_event() -> None:
    """Test synchronous event reading."""
    header = {
        "type": "test-event",
        "version": wyoming_version,
        "data_length": len(DATA_BYTES),
        "payload_length": len(PAYLOAD),
        # inline data
        "data": {
            "test": "this will be overwritten by DATA",
            "test2": "this will not",
        },
    }

    with io.BytesIO() as buf:
        # First line is JSON with event type and data/payload lengths.
        header_bytes = json.dumps(header, ensure_ascii=False).encode("utf-8")
        buf.write(header_bytes)
        buf.write(b"\n")

        # Data dict comes next, encoded as UTF-8 JSON.
        buf.write(DATA_BYTES)

        # Payload comes last
        buf.write(PAYLOAD)

        buf.seek(0)
        event_bytes = buf.getvalue()

    with io.BytesIO(event_bytes) as reader:
        event = read_event(reader)
        assert event == Event(
            type="test-event",
            # inline data was overwritten
            data={"test": "data", "test2": "this will not"},
            payload=PAYLOAD,
        )


def test_read_event_no_data() -> None:
    """Test synchronous event reading with no data."""
    header = {"type": "test-event", "version": wyoming_version}

    with io.BytesIO() as buf:
        header_bytes = json.dumps(header, ensure_ascii=False).encode("utf-8")
        buf.write(header_bytes)
        buf.write(b"\n")
        buf.seek(0)
        event_bytes = buf.getvalue()

    with io.BytesIO(event_bytes) as reader:
        event = read_event(reader)
        assert event == Event(type="test-event")


@pytest.mark.asyncio
async def test_async_read_event() -> None:
    """Test asynchronous event reading."""
    header = {
        "type": "test-event",
        "version": wyoming_version,
        "data_length": len(DATA_BYTES),
        "payload_length": len(PAYLOAD),
        # inline data
        "data": {
            "test": "this will be overwritten by DATA",
            "test2": "this will not",
        },
    }

    with io.BytesIO() as buf:
        # First line is JSON with event type and data/payload lengths.
        header_bytes = json.dumps(header, ensure_ascii=False).encode("utf-8")
        buf.write(header_bytes)
        buf.write(b"\n")

        # Data dict comes next, encoded as UTF-8 JSON.
        buf.write(DATA_BYTES)

        # Payload comes last
        buf.write(PAYLOAD)

        buf.seek(0)
        event_bytes = buf.getvalue()

    reader = FakeStreamReader(event_bytes)
    event = await async_read_event(reader)  # type: ignore
    assert event == Event(
        type="test-event",
        # inline data was overwritten
        data={"test": "data", "test2": "this will not"},
        payload=PAYLOAD,
    )


@pytest.mark.asyncio
async def test_async_read_event_no_data() -> None:
    """Test asynchronous event reading with no data."""
    header = {"type": "test-event", "version": wyoming_version}

    with io.BytesIO() as buf:
        header_bytes = json.dumps(header, ensure_ascii=False).encode("utf-8")
        buf.write(header_bytes)
        buf.write(b"\n")
        buf.seek(0)
        event_bytes = buf.getvalue()

    reader = FakeStreamReader(event_bytes)
    event = await async_read_event(reader)  # type: ignore
    assert event == Event(type="test-event")



import asyncio
import sys
from abc import ABC, abstractmethod
from functools import partial
from pathlib import Path
from typing import Callable, Dict, Optional, Union
from urllib.parse import urlparse

from .event import Event, async_get_stdin, async_read_event, async_write_event


class AsyncEventHandler(ABC):
    """Base class for async Wyoming event handler."""

    def __init__(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        self.reader = reader
        self.writer = writer
        self._is_running = False

    @abstractmethod
    async def handle_event(self, event: Event) -> bool:
        """Handle an event. Returning false will disconnect the client."""
        return True

    async def write_event(self, event: Event) -> None:
        """Send an event to the client."""
        await async_write_event(event, self.writer)

    async def run(self) -> None:
        """Receive events until stopped or handle_event returns false."""
        self._is_running = True

        try:
            while self._is_running:
                event = await async_read_event(self.reader)
                if event is None:
                    break

                if not (await self.handle_event(event)):
                    break
        finally:
            await self.disconnect()
            self.writer.close()

    async def disconnect(self) -> None:
        """Called when client disconnects."""

    async def stop(self) -> None:
        """Try to stop the event handler."""
        self._is_running = False
        self.writer.close()
        self.reader.feed_eof()


HandlerFactory = Callable[
    [asyncio.StreamReader, asyncio.StreamWriter], AsyncEventHandler
]


class AsyncServer(ABC):
    """Base class for async Wyoming server."""

    def __init__(self) -> None:
        self._handlers: Dict[asyncio.Task, AsyncEventHandler] = {}

    @abstractmethod
    async def run(self, handler_factory: HandlerFactory) -> None:
        """Start server and block while running."""

    @staticmethod
    def from_uri(uri: str) -> "AsyncServer":
        """Create server from URI."""
        result = urlparse(uri)

        if result.scheme == "unix":
            return AsyncUnixServer(result.path)

        if result.scheme == "tcp":
            if (result.hostname is None) or (result.port is None):
                raise ValueError("A port must be specified when using a 'tcp://' URI")

            return AsyncTcpServer(result.hostname, result.port)

        if result.scheme == "stdio":
            return AsyncStdioServer()

        raise ValueError("Only 'stdio://', 'unix://', or 'tcp://' are supported")

    async def _handler_callback(
        self,
        handler_factory: HandlerFactory,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ):
        handler = handler_factory(reader, writer)
        task = asyncio.create_task(handler.run(), name="wyoming event handler")
        self._handlers[task] = handler
        task.add_done_callback(lambda t: self._handlers.pop(t, None))

    async def start(self, handler_factory: HandlerFactory) -> None:
        """Start server without blocking."""

    async def stop(self) -> None:
        """Try to stop all event handlers."""
        await asyncio.gather(*(h.stop() for h in self._handlers.values()))


class AsyncStdioServer(AsyncServer):
    """Wyoming server over stdin/stdout."""

    async def run(self, handler_factory: HandlerFactory) -> None:
        """Start server and block while running."""
        reader = await async_get_stdin()

        # Get stdout writer.
        # NOTE: This will make print() non-blocking.
        loop = asyncio.get_running_loop()
        writer_transport, writer_protocol = await loop.connect_write_pipe(
            asyncio.streams.FlowControlMixin, sys.stdout
        )
        writer = asyncio.StreamWriter(writer_transport, writer_protocol, None, loop)

        handler = handler_factory(reader, writer)
        while True:
            event = await async_read_event(reader)
            if event is None:
                break

            if not (await handler.handle_event(event)):
                break


class AsyncTcpServer(AsyncServer):
    """Wyoming server over TCP."""

    def __init__(self, host: str, port: int) -> None:
        super().__init__()
        self.host = host
        self.port = port
        self._server: Optional[asyncio.AbstractServer] = None

    async def run(self, handler_factory: HandlerFactory) -> None:
        handler_callback = partial(self._handler_callback, handler_factory)
        self._server = await asyncio.start_server(
            handler_callback, host=self.host, port=self.port
        )

        await self._server.serve_forever()

    async def start(self, handler_factory: HandlerFactory) -> None:
        """Start server without blocking."""
        handler_callback = partial(self._handler_callback, handler_factory)
        self._server = await asyncio.start_server(
            handler_callback, host=self.host, port=self.port
        )

        await self._server.start_serving()

    async def stop(self) -> None:
        """Try to stop all event handlers."""
        await super().stop()

        if self._server is not None:
            self._server.close()


class AsyncUnixServer(AsyncServer):
    """Wyoming server over a Unix domain socket."""

    def __init__(self, socket_path: Union[str, Path]) -> None:
        super().__init__()
        self.socket_path = Path(socket_path)
        self._server: Optional[asyncio.AbstractServer] = None

    async def run(self, handler_factory: HandlerFactory) -> None:
        """Start server and block while running."""
        # Need to unlink socket file if it exists
        self.socket_path.unlink(missing_ok=True)

        handler_callback = partial(self._handler_callback, handler_factory)
        self._server = await asyncio.start_unix_server(
            handler_callback, path=self.socket_path
        )

        try:
            await self._server.serve_forever()
        finally:
            # Unlink when we're done
            self.socket_path.unlink(missing_ok=True)

    async def start(self, handler_factory: HandlerFactory) -> None:
        """Start server without blocking."""
        # Need to unlink socket file if it exists
        self.socket_path.unlink(missing_ok=True)

        handler_callback = partial(self._handler_callback, handler_factory)
        self._server = await asyncio.start_unix_server(
            handler_callback, path=self.socket_path
        )

        await self._server.start_serving()

    async def stop(self) -> None:
        """Try to stop all event handlers."""
        await super().stop()

        if self._server is not None:
            self._server.close()

        self.socket_path.unlink(missing_ok=True)



"""Support for voice timers."""

from dataclasses import dataclass
from typing import Optional

from .event import Event, Eventable

DOMAIN = "timer"
_STARTED_TYPE = "timer-started"
_UPDATED_TYPE = "timer-updated"
_CANCELLED_TYPE = "timer-cancelled"
_FINISHED_TYPE = "timer-finished"


@dataclass
class TimerStarted(Eventable):
    """New timer was started."""

    id: str
    """Unique id of timer."""

    total_seconds: int
    """Total number of seconds the timer will run for."""

    name: Optional[str] = None
    """Optional name provided by user."""

    start_hours: Optional[int] = None
    """Number of hours users requested the timer to run for."""

    start_minutes: Optional[int] = None
    """Number of minutes users requested the timer to run for."""

    start_seconds: Optional[int] = None
    """Number of minutes users requested the timer to run for."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _STARTED_TYPE

    def event(self) -> Event:
        data = {"id": self.id, "total_seconds": self.total_seconds}
        if self.name is not None:
            data["name"] = self.name

        if self.start_hours is not None:
            data["start_hours"] = self.start_hours

        if self.start_minutes is not None:
            data["start_minutes"] = self.start_minutes

        if self.start_seconds is not None:
            data["start_seconds"] = self.start_seconds

        return Event(
            type=_STARTED_TYPE,
            data=data,
        )

    @staticmethod
    def from_event(event: Event) -> "TimerStarted":
        return TimerStarted(
            id=event.data["id"],
            total_seconds=event.data["total_seconds"],
            name=event.data.get("name"),
            start_hours=event.data.get("start_hours"),
            start_minutes=event.data.get("start_minutes"),
            start_seconds=event.data.get("start_seconds"),
        )


@dataclass
class TimerUpdated(Eventable):
    """Existing timer was paused, resumed, or had time added or removed."""

    id: str
    """Unique id of timer."""

    is_active: bool
    """True if timer is running."""

    total_seconds: int
    """Number of seconds left on the timer."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _UPDATED_TYPE

    def event(self) -> Event:
        return Event(
            type=_UPDATED_TYPE,
            data={
                "id": self.id,
                "is_active": self.is_active,
                "total_seconds": self.total_seconds,
            },
        )

    @staticmethod
    def from_event(event: Event) -> "TimerUpdated":
        return TimerUpdated(
            id=event.data["id"],
            is_active=event.data["is_active"],
            total_seconds=event.data["total_seconds"],
        )


@dataclass
class TimerCancelled(Eventable):
    """Existing timer was cancelled."""

    id: str
    """Unique id of timer."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _CANCELLED_TYPE

    def event(self) -> Event:
        return Event(
            type=_CANCELLED_TYPE,
            data={"id": self.id},
        )

    @staticmethod
    def from_event(event: Event) -> "TimerCancelled":
        return TimerCancelled(id=event.data["id"])


@dataclass
class TimerFinished(Eventable):
    """Existing timer finished without being cancelled."""

    id: str
    """Unique id of timer."""

    @staticmethod
    def is_type(event_type: str) -> bool:
        return event_type == _FINISHED_TYPE

    def event(self) -> Event:
        return Event(
            type=_FINISHED_TYPE,
            data={"id": self.id},
        )

    @staticmethod
    def from_event(event: Event) -> "TimerFinished":
        return TimerFinished(id=event.data["id"])



"""HTTP server for text to speech (TTS)."""

import io
import logging
import wave
from pathlib import Path
from typing import Optional

from flask import Response, request

from wyoming.audio import AudioChunk, AudioStart, AudioStop
from wyoming.client import AsyncClient
from wyoming.error import Error
from wyoming.tts import Synthesize, SynthesizeVoice

from .shared import get_app, get_argument_parser

_DIR = Path(__file__).parent
CONF_PATH = _DIR / "conf" / "tts.yaml"


def main():
    parser = get_argument_parser()
    parser.add_argument("--voice", help="Default voice for synthesis")
    parser.add_argument("--speaker", help="Default voice speaker for synthesis")
    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    app = get_app("tts", CONF_PATH, args)

    @app.route("/api/text-to-speech", methods=["POST", "GET"])
    async def api_stt() -> Response:
        uri = request.args.get("uri", args.uri)
        if not uri:
            raise ValueError("URI is required")

        if request.method == "POST":
            text = request.data.decode()
        else:
            text = request.args.get("text", "")

        if not text:
            raise ValueError("Text is required")

        voice: Optional[SynthesizeVoice] = None
        voice_name = request.args.get("voice", args.voice)
        if voice_name:
            voice = SynthesizeVoice(
                name=voice_name, speaker=request.args.get("speaker", args.speaker)
            )

        async with AsyncClient.from_uri(uri) as client:
            wav_io = io.BytesIO()
            wav_file = wave.open(wav_io, "wb")
            await client.write_event(Synthesize(text=text, voice=voice).event())

            while True:
                event = await client.read_event()
                if event is None:
                    raise RuntimeError("Client disconnected")

                if AudioStart.is_type(event.type):
                    audio_start = AudioStart.from_event(event)
                    wav_file.setframerate(audio_start.rate)
                    wav_file.setsampwidth(audio_start.width)
                    wav_file.setnchannels(audio_start.channels)
                elif AudioChunk.is_type(event.type):
                    audio_chunk = AudioChunk.from_event(event)
                    wav_file.writeframes(audio_chunk.audio)
                elif AudioStop.is_type(event.type):
                    wav_file.close()
                    wav_io.seek(0)
                    return Response(
                        wav_io.getvalue(), headers={"Content-Type": "audio/wav"}
                    )
                elif Error.is_type(event.type):
                    error = Error.from_event(event)
                    raise RuntimeError(
                        f"Unexpected error from client: code={error.code}, text={error.text}"
                    )

    app.run(args.host, args.port)


if __name__ == "__main__":
    main()



import asyncio
from abc import ABC
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlparse

from .event import (
    Event,
    async_get_stdin,
    async_get_stdout,
    async_read_event,
    async_write_event,
)


class AsyncClient(ABC):
    """Base class for Wyoming async client."""

    def __init__(self) -> None:
        self._reader: Optional[asyncio.StreamReader] = None
        self._writer: Optional[asyncio.StreamWriter] = None

    async def read_event(self) -> Optional[Event]:
        assert self._reader is not None
        return await async_read_event(self._reader)

    async def write_event(self, event: Event) -> None:
        assert self._writer is not None
        await async_write_event(event, self._writer)

    async def connect(self) -> None:
        pass

    async def __aenter__(self):
        await self.connect()
        return self

    async def disconnect(self) -> None:
        pass

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.disconnect()

    @staticmethod
    def from_uri(uri: str) -> "AsyncClient":
        result = urlparse(uri)

        if result.scheme == "unix":
            return AsyncUnixClient(result.path)

        if result.scheme == "tcp":
            if (result.hostname is None) or (result.port is None):
                raise ValueError("A port must be specified when using a 'tcp://' URI")

            return AsyncTcpClient(result.hostname, result.port)

        if result.scheme == "stdio":
            return AsyncStdioClient()

        raise ValueError("Only 'stdio://', 'unix://', or 'tcp://' are supported")


class AsyncTcpClient(AsyncClient):
    """TCP Wyoming client."""

    def __init__(self, host: str, port: int) -> None:
        super().__init__()

        self.host = host
        self.port = port

    async def connect(self) -> None:
        self._reader, self._writer = await asyncio.open_connection(
            host=self.host,
            port=self.port,
        )

    async def disconnect(self) -> None:
        writer = self._writer
        self._reader = None
        self._writer = None

        if writer is not None:
            writer.close()
            await writer.wait_closed()


class AsyncUnixClient(AsyncClient):
    """Unix domain socket Wyoming client."""

    def __init__(self, socket_path: Union[str, Path]) -> None:
        super().__init__()

        self.socket_path = Path(socket_path)

    async def connect(self) -> None:
        self._reader, self._writer = await asyncio.open_unix_connection(
            path=self.socket_path
        )

    async def disconnect(self) -> None:
        writer = self._writer
        self._reader = None
        self._writer = None

        if writer is not None:
            writer.close()
            await writer.wait_closed()


class AsyncStdioClient(AsyncClient):
    """Standard output Wyoming client."""

    def __init__(self) -> None:
        super().__init__()

        self._reader: Optional[asyncio.StreamReader] = None
        self._writer: Optional[asyncio.StreamWriter] = None

    async def read_event(self) -> Optional[Event]:
        if self._reader is None:
            self._reader = await async_get_stdin()

        assert self._reader is not None
        return await async_read_event(self._reader)

    async def write_event(self, event: Event) -> None:
        if self._writer is None:
            self._writer = await async_get_stdout()

        assert self._writer is not None
        await async_write_event(event, self._writer)



"""HTTP server for automated speech recognition (ASR)."""

import io
import logging
import wave
from pathlib import Path

from flask import Response, jsonify, request

from wyoming.asr import Transcribe, Transcript
from wyoming.audio import wav_to_chunks
from wyoming.client import AsyncClient
from wyoming.error import Error

from .shared import get_app, get_argument_parser

_DIR = Path(__file__).parent
CONF_PATH = _DIR / "conf" / "asr.yaml"


def main():
    parser = get_argument_parser()
    parser.add_argument("--model", help="Default model name for transcription")
    parser.add_argument("--language", help="Default language for transcription")
    parser.add_argument("--samples-per-chunk", type=int, default=1024)
    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    app = get_app("asr", CONF_PATH, args)

    @app.route("/api/speech-to-text", methods=["POST"])
    async def api_stt() -> Response:
        uri = request.args.get("uri", args.uri)
        if not uri:
            raise ValueError("URI is required")

        model_name = request.args.get("model", args.model)
        language = request.args.get("language", args.model)

        async with AsyncClient.from_uri(uri) as client:
            await client.write_event(
                Transcribe(name=model_name, language=language).event()
            )

            with io.BytesIO(request.data) as wav_io:
                with wave.open(wav_io, "rb") as wav_file:
                    chunks = wav_to_chunks(
                        wav_file,
                        samples_per_chunk=args.samples_per_chunk,
                        start_event=True,
                        stop_event=True,
                    )
                    for chunk in chunks:
                        await client.write_event(chunk.event())

            while True:
                event = await client.read_event()
                if event is None:
                    raise RuntimeError("Client disconnected")

                if Transcript.is_type(event.type):
                    transcript = Transcript.from_event(event)
                    return jsonify(transcript.to_dict())

                if Error.is_type(event.type):
                    error = Error.from_event(event)
                    raise RuntimeError(
                        f"Unexpected error from client: code={error.code}, text={error.text}"
                    )

    app.run(args.host, args.port)


if __name__ == "__main__":
    main()
